# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import Union

Flint = Union[int, float]  # float or int
Paths = Iterable[str]  # iterable of path-strings
JASM = Union[None, bool, int, float, str, List[Any], Dict[str, Any]]  # JSON obj



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

from typing import Any
from typing import Optional

def timestamp(ts: Optional[float] = ...) -> str: ...
def environ_dict(): ...
def linked_tmp_dir(
    suffix: Optional[Any] = ...,
    prefix: Optional[Any] = ...,
    dir: Optional[Any] = ...,
    mkdirs: Any = ...,
    lndirs: Any = ...,
    lnfiles: Any = ...,
) -> None: ...
def prinfo(obj: Any) -> None: ...
def pyfilepath(split: bool = ...): ...
def time_funk(funk: Any, *args: Any, **kwargs: Any): ...
def cmp_funks(f1: Any, f2: Any, runs: int = ..., *args: Any, **kwargs: Any): ...



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python
from cProfile import Profile
from functools import wraps
from inspect import getfile
from logging import DEBUG
from logging import config
from logging import getLogger
from os import chdir
from os import getcwd
from os import makedirs
from os import mkdir
from os import path
from pstats import SortKey
from time import time

from pupy._jasm import json
from pupy.fmt import nseconds

# "%(color)s[%(levelname)1.1s %(asctime)s %(module)s:%(lineno)d]%(end_color)s %(message)s"
logging_config = dict(
    version=1,
    formatters={
        "f": {
            "format": "[%(levelname)1.1s %(asctime)s %(filename)s:%(lineno)d] %(message)s"
        }
    },
    handlers={
        "h": {"class": "logging.StreamHandler", "formatter": "f", "level": DEBUG}
    },
    root={"handlers": ["h"], "level": DEBUG},
)

config.dictConfig(logging_config)

logger = getLogger()


def in_n_out(funk):
    """Chdir in to the dir the test_function is in and change dirs out when done

    :type funk: Callable
    :param funk: docin.api logger functions logger.(debug/info/warn/error)
    :return: wrapped function
    """

    @wraps(funk)
    def chin_n_chout(*args, **kwargs):
        """

        :param args:
        :param kwargs:
        :return:
        """
        # thing = os.path.split(os.path.realpath(__file__))
        cd = getcwd()
        dirpath = args[0]
        if path.isdir(dirpath):
            chdir(dirpath)
        funk_results = funk(*args, **kwargs)
        chdir(cd)
        return funk_results

    return chin_n_chout


def flog(funk=None, loglevel="debug", funk_call=True, tictoc=False):
    """

    :param funk:
    :param loglevel:
    :param funk_call:
    :param tictoc:
    :return:
    """
    _log_levels = {
        "debug": logger.debug,
        "info": logger.info,
        "warn": logger.warning,
        "error": logger.error,
    }

    def _decorate_flog_wrapper(_funk):
        def _fmt_args(*args):
            return ", ".join(str(arg) for arg in args)

        def _fmt_kwargs(**kwargs):
            return ", ".join("{}={}".format(str(k), str(v)) for k, v in kwargs.items())

        def _fmt_call(*args, **kwargs):
            params_str = ", ".join(
                s for s in (_fmt_args(*args), _fmt_kwargs(**kwargs)) if s
            )
            return "{}({})".format(_funk.__name__, params_str)

        @wraps(_funk)
        def _flog_wrapper(*args, **kwargs):
            ti = time()
            _ret = _funk(*args, **kwargs)
            tf = time()
            msg_parts = [
                _fmt_call(*args, **kwargs) if funk_call else None,
                nseconds(ti, tf) if tictoc else None,
            ]
            msg_str = " | ".join(part for part in msg_parts if part)
            if any(el for el in msg_parts):
                _log_levels[loglevel]("[FLOG] | {}".format(msg_str))
            return _ret

        return _flog_wrapper

    return _decorate_flog_wrapper(funk) if funk else _decorate_flog_wrapper


def dirdec(funk):
    """

    :param funk:
    :return:
    """

    @wraps(funk)
    def _wrapper(*args, **kwargs):
        result = funk(*args, **kwargs)
        try:
            mkdir(result)
        except (FileExistsError, OSError) as e:
            pass
        return result

    return _wrapper


def mkdirs(funk):
    """

    :param funk:
    :return:
    """

    @wraps(funk)
    def _wrapper(*args, **kwargs):
        try:
            dirpath = path.split(args[0])[0]
        except IndexError:
            dirpath = path.split(kwargs["filepath"])[0]

        try:
            makedirs(dirpath, exist_ok=True)
        except OSError as e:
            # print(e)
            pass
        except TypeError as e:
            # print(e)
            pass
        return funk(*args, **kwargs)

    return _wrapper


def cash_it(funk):
    """args-2-return value cache.

    This function is particularly useful for when you want that lru-cache,
    but you/one is working with python two.

    :param funk: function to be cached
    :type funk: function
    :returns: wrapped function
    :rtype: function

    """
    cash_money = {}

    @wraps(funk)
    def cash_wrap(*argz):
        """

        """
        if argz not in cash_money:
            rv = funk(*argz)
            cash_money[argz] = rv
        return cash_money[argz]

    return cash_wrap


def cprof(funk):
    """"cProfiling decorator

    src: https://zapier.com/engineering/profiling-python-boss/

    :param funk: funktion to decorate and 'get tha c prof of'

    """

    @wraps(funk)
    def profiled_funk(*args, **kwargs):
        """wrapper funk"""

        profile = Profile()
        try:
            profile.enable()
            ret_val = funk(*args, **kwargs)
            profile.disable()
        finally:
            print("__CPROFILE__")
            profile.print_stats(SortKey.CUMULATIVE)
        return ret_val

    return profiled_funk


class tictoc(object):
    """Timing decorator object

    :param runs: # of runs to time over (defaults to 1)

    """

    def __init__(self, runs=1, show_args=False, indent=True):
        self.runs = runs
        self.show_args = show_args
        self.indent = indent

    def _old_fmt(self, t_total, funk, args_string):
        _fmt_strs = (
            "__TICTOC__",
            "    file: {}".format(getfile(funk)),
            "    funk: {}".format(funk.__name__),
            "    args: {}".format(str(args_string)[:24]) if self.show_args else None,
            "    time: {}".format(nseconds(t_total)),
            "    runs: {}".format(self.runs),
        )

        return "\n".join(filter(None, _fmt_strs))

    def __str__(self, t_total, funk, args_string):
        d = {
            "file": getfile(funk),
            "funk": funk.__name__,
            "args": str(args_string)[:24] if self.show_args else None,
            "time": nseconds(t_total),
            "runs": self.runs,
        }
        d = {k: v for k, v in d.items() if v}
        return "{} {}".format(
            "__TICTOC__:", json.dumps(d, indent=4 if self.indent else None)
        )

    def __call__(self, time_funk, printing=True):
        @wraps(time_funk)
        def time_wrapper(*args, **kwargs):
            """

            :param args:
            :param kwargs:
            :return:
            """
            self.args = str(args)
            ts = time()
            for i in range(self.runs):
                result = time_funk(*args, **kwargs)
            te = time()
            t_total = (te - ts) / self.runs
            try:
                if printing:
                    print(self.__str__(t_total, time_funk, self.args))
            except Exception as e:
                pass
            return result

        return time_wrapper


def prop(fn):
    """Lazy property decorator"""
    attr_name = "_prop_" + fn.__name__

    @property
    def _prop(self):
        if not hasattr(self, attr_name):
            setattr(self, attr_name, fn(self))
        return getattr(self, attr_name)

    return _prop


def requires(package):
    def _requires(_funk):
        @wraps(_funk)
        def _wrapper(*args, **kwargs):
            try:
                return _funk(*args, **kwargs)
            except ImportError as e:
                raise ImportError("'pip install {}' to use this!".format(package))

        return _wrapper

    return _requires



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

from typing import Any
from typing import Iterator
from typing import List
from typing import Optional

from pupy._typing import Flint

def nbytes(num: Flint) -> str: ...
def filesize(filepath: str) -> str: ...
def nseconds(t1: float, t2: Optional[float] = ...) -> str: ...
def term_table(
    strings: List[str], row_wise: bool = ..., filler: str = ...
) -> Iterator[Any]: ...
def bytes2str(bites: bytes, encoding: str = ...) -> str: ...
def binary_string(number: int) -> str: ...
def strip_comments(string: str) -> str: ...
def strip_ascii(s: str) -> str: ...
def no_b(string: str) -> str: ...
def no_u(string: str) -> str: ...
def rhex_str(length: int = ...) -> str: ...



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python
from pupy import savings_n_loads as _io
from pupy._alias import p
from pupy._alias import pp
from pupy.decorations import cash_it
from pupy.decorations import dirdec
from pupy.decorations import mkdirs
from pupy.decorations import prop
from pupy.decorations import tictoc
from pupy.foreign import chunks
from pupy.foreign import digits_list
from pupy.foreign import dirs_gen
from pupy.foreign import exhaust
from pupy.foreign import files_dirs_gen
from pupy.foreign import files_gen
from pupy.foreign import int_from_digits
from pupy.foreign import is_permutation
from pupy.foreign import iter_product
from pupy.foreign import rotate
from pupy.foreign import rotations_gen
from pupy.foreign import spliterable
from pupy.foreign import walk_gen
from pupy.savings_n_loads import ljson
from pupy.savings_n_loads import load_jasm
from pupy.savings_n_loads import lpak
from pupy.savings_n_loads import lstr
from pupy.savings_n_loads import lstring
from pupy.savings_n_loads import safepath
from pupy.savings_n_loads import save_jasm
from pupy.savings_n_loads import savings
from pupy.savings_n_loads import sjson
from pupy.savings_n_loads import spak
from pupy.savings_n_loads import sstr
from pupy.savings_n_loads import sstring
from pupy.sh import basename
from pupy.sh import cd
from pupy.sh import cp
from pupy.sh import dirname
from pupy.sh import echo
from pupy.sh import ls
from pupy.sh import mv
from pupy.sh import parent_dirpath
from pupy.sh import pwd
from pupy.sh import rm
from pupy.utils import prinfo
from pupy.utils import pyfilepath
from pupy.utils import timestamp

__all__ = [e for e in dir() if not e.startswith("_")]



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

from typing import Any
from typing import Optional

logging_config: Any
logger: Any

def in_n_out(funk: Any): ...
def flog(
    funk: Optional[Any] = ...,
    loglevel: str = ...,
    funk_call: bool = ...,
    tictoc: bool = ...,
): ...
def dirdec(funk: Any): ...
def mkdirs(funk: Any): ...
def cash_it(funk: Any): ...
def cprof(funk: Any): ...

class tictoc:
    runs: Any = ...
    show_args: Any = ...
    indent: Any = ...
    def __init__(
        self, runs: int = ..., show_args: bool = ..., indent: bool = ...
    ) -> None: ...
    args: Any = ...
    def __call__(self, time_funk: Any, printing: bool = ...): ...

def prop(fn: Any): ...
def requires(package: Any): ...



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

from collections.abc import MutableSequence
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Union

def prime_gen(
    plim: int = ..., kprimes: Union[None, Iterable[int]] = ...
) -> Iterator[int]: ...
def prime_factorization_gen(n: int) -> Iterator[int]: ...
def prime_factors_gen(n: int) -> Iterator[Any]: ...
def is_prime(number: int) -> bool: ...

class OctopusPrime(MutableSequence):
    max_loaded: Any = ...
    def __init__(self, plim: int = ...) -> None: ...
    def primes_below(self, upper_bound: Any): ...
    def primes_between(self, lower_bound: int, upper_bound: int) -> List[int]: ...
    def __len__(self) -> int: ...
    def __getitem__(self, i: Union[int, slice]) -> Union[int, List[int]]: ...
    def __delitem__(self, i: Any) -> None: ...
    def __setitem__(self, key: Any, value: Any) -> None: ...
    def insert(self, index: Any, object: Any) -> None: ...



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python
"""
===============
Savings & Loads
===============

Common funks for ur everyday savings and loads. 

If you are a friend of Jasm/Json/Jason-Greenberg like myself, 
'pip install ujson' for a better time.

If you need it... you can always...
pip install (msgpack, toml, ruamel.yaml) for the whole shebang.

"""

from codecs import getwriter
from io import open
from itertools import count
from os import path
from os import utime
from typing import Union

from pupy._jasm import json
from pupy._typing import JASM
from pupy.decorations import mkdirs

try:
    from toml import dumps as toml_dumps
    from toml import load as toml_load
    from toml import loads as toml_loads
except ModuleNotFoundError:
    pass

try:
    from msgpack import pack
    from msgpack import unpack
except ModuleNotFoundError:
    pass

try:
    from ruamel.yaml import YAML

    _yaml_saver = YAML()
    _yaml_loader = YAML(typ="safe")
except ModuleNotFoundError:
    pass


def safepath(path_str: str) -> str:
    """Checks if a file/dir path is save/unused; returns an unused path.

    :param path_str: file or dir path
    :return: A file/dir path that does not exist and contains the given path

    """
    if path.exists(path_str):
        f_bn, f_ext = path.splitext(path_str)
        for n in count(1):
            safe_save_path = f_bn + "_({}).".format(str(n)) + f_ext
            if not path.exists(safe_save_path):
                return safe_save_path
    return path_str


def lbytes(filepath: str) -> Union[bytes, bytearray]:
    """Read bytes from file path

    :param filepath: filepath as as string to read bites from
    :return: some bytes...

    """
    with open(filepath, "rb") as file:
        return file.read()


def sbytes(filepath: str, bytes: Union[bytes, bytearray]) -> None:
    """Write bytes to a filepath

    :param filepath: filepath as as string to read bites from
    :param bytes:
    :return: some bytes...

    """
    with open(filepath, "wb") as file:
        return file.write(bytes)


@mkdirs
def sstring(filepath: str, string: str) -> None:
    """Writes a string to filepath


    :param filepath: Filepath save location
    :param string: File as a string to be saved
    :return: None? what do you want? confirmation?

    .. note:: Decorated w/ @mkdirs
        Function is decorated with @mkdirs decorator; @mkdirs creates parent
        directories for the given filepath if they do not already exist.

    """
    with open(filepath, "wb") as file:
        file.write(string.encode("utf-8"))


def savings(filepath: str, string: str) -> None:
    """Alias for sstring"""
    return sstring(filepath, string)


def sstr(filepath: str, string: str) -> None:
    """Alias for sstring"""
    return sstring(filepath, string)


def lstring(filepath: str) -> str:
    """(lstring) Read and return the file-contents as a string given a filepath

    :param filepath: Path to a file to read
    :return: Content of the file read as a string

    """
    try:
        with open(filepath, "r", encoding="utf-8") as f:
            return f.read()
    except UnicodeDecodeError:
        with open(filepath, "r", encoding="latin2") as f:
            return f.read()


def lstr(filepath: str) -> str:
    """Alias for lstring"""
    return lstring(filepath)


@mkdirs
def sjson(filepath: str, data: JASM, minify: bool = False) -> None:
    """Save json-serial-ize-able data to a specific filepath.

    :param filepath: destination filepath
    :param data: json cereal-izable dictionary/list/thing
    :param minify: Bool flag -- minify the json file
    :return: None

    """
    if isinstance(data, dict) and any(isinstance(val, bytes) for val in data.values()):
        data = {k: str(v, encoding="utf-8") for k, v in data.items()}
    if minify:
        json_string = json.dumps(data, ensure_ascii=True)
    else:
        json_string = json.dumps(data, indent=4, sort_keys=True, ensure_ascii=True)
    sstring(filepath, json_string)


def save_jasm(filepath: str, data: JASM, minify: bool = False) -> None:
    """Alias for sjson (which stands for 'save-json')"""
    return sjson(filepath, data, minify)


def sjasm(filepath: str, data: JASM, minify: bool = False) -> None:
    """Alias for sjson (which stands for 'save-json')"""
    return sjson(filepath, data, minify)


def ljson(filepath: str) -> JASM:
    """Load a json file given a filepath and return the file-data

    :param filepath: path to the jasm file you want to load
    :return: Loaded file contents
    """
    try:
        with open(filepath) as infile:
            return json.load(infile)
    except UnicodeDecodeError as e:
        return json.loads(lstring(filepath))


def load_jasm(filepath: str) -> JASM:
    """Alias for ljson (which stands for 'load-json')"""
    return ljson(filepath)


def ljasm(filepath: str) -> JASM:
    """Alias for ljson (which stands for 'load-json')"""
    return ljson(filepath)


@mkdirs
def touch(filepath: str) -> None:
    """Touches a file just like touch on the command line

    :param filepath: filepath to 'touch' in a unix-y sense
    :return: None
    """
    with open(filepath, "a"):
        utime(filepath, None)


def shebang(filepath: str) -> Union[None, str]:
    """returns the shebang path given a filepath or None if it does not exist.

    :param filepath: path to a file w/ a shebange line
    :return: shebang line or None

    .. doctest::python

        >>> from inspect import getabsfile
        >>> script = 'ashellscript.sh'
        >>> with open(script, 'w') as f:
        ...     f.write('#!/bin/bash\\necho "howdy"\\n')
        25
        >>> shebang(script)
        '#!/bin/bash'
        >>> from os import remove
        >>> remove(script)

    """
    with open(filepath, "r") as f:
        first = f.readline().strip("\n")
        return first if "#!" in first[:2] else None


def stoml(filepath: str, data: JASM) -> None:
    try:
        filepath = filepath if "." in filepath else "{}.toml".format(filepath)
        sstring(filepath, toml_dumps(data))
        return filepath
    except NameError:
        raise EnvironmentError("'pip install toml' if you wanna use this!")


def ltoml(filepath: str) -> JASM:
    try:
        with open(filepath) as f:
            return toml_load(f)
    except NameError:
        raise EnvironmentError("'pip install toml' if you wanna use this!")
    except UnicodeDecodeError as e:
        return toml_loads(lstring(filepath))


def spak(filepath: str, data: JASM) -> None:
    try:
        filepath = filepath if "." in filepath else "{}.pak".format(filepath)
        with open(filepath, "wb") as outfile:
            pack(data, outfile)
        return filepath
    except NameError:
        raise EnvironmentError("'pip install msgpack' if you wanna use this!")


def lpak(filepath: str, raw: bool = False) -> JASM:
    try:
        with open(filepath, "rb") as data_file:
            return unpack(data_file, raw=raw)
    except NameError:
        raise EnvironmentError("'pip install msgpack' if you wanna use this!")


def syaml(filepath: str, data: JASM) -> None:
    try:
        filepath = filepath if "." in filepath else "{}.yml".format(filepath)
        with open(filepath, "w") as data_file:
            _yaml_saver.dump(data, data_file)
        return filepath
    except NameError:
        raise EnvironmentError("'pip install ruamel.yaml' if you wanna use this!")


def lyaml(filepath: str) -> JASM:
    try:
        with open(filepath) as data_file:
            return _yaml_loader.load(data_file)
    except NameError:
        raise EnvironmentError("'pip install ruamel.yaml' if you wanna use this!")



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

from os import chdir
from os import getcwd
from typing import Any
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

class LIN:
    @staticmethod
    def rsync_args(
        src: Any,
        dest: Any,
        delete: bool = ...,
        mkdirs: bool = ...,
        exclude: Any = ...,
        include: Any = ...,
        dry_run: bool = ...,
    ): ...
    @staticmethod
    def rsync(
        src: Any,
        dest: Any,
        delete: bool = ...,
        exclude: Any = ...,
        include: Any = ...,
        dry_run: bool = ...,
    ): ...
    @staticmethod
    def link_dir(linkpath: Any, targetpath: Any) -> None: ...
    @staticmethod
    def link_dirs(link_target_tuples: Any) -> None: ...
    @staticmethod
    def link_file(linkpath: str, targetpath: str) -> None: ...
    @staticmethod
    def link_files(link_target_tuples: Any) -> None: ...
    @staticmethod
    def unlink_dir(link: Any) -> None: ...
    @staticmethod
    def unlink_dirs(links: Any) -> None: ...
    @staticmethod
    def unlink_file(link: Any) -> None: ...
    @staticmethod
    def unlink_files(links: Any) -> None: ...
    sync: Any = ...

class WIN:
    @staticmethod
    def robocopy_args(
        src: Any,
        dest: Any,
        delete: bool = ...,
        exclude_files: Any = ...,
        exclude_dirs: Any = ...,
        dry_run: bool = ...,
    ): ...
    @staticmethod
    def robocopy(
        src: Any,
        dest: Any,
        delete: bool = ...,
        exclude_files: Any = ...,
        exclude_dirs: Any = ...,
        dry_run: bool = ...,
    ): ...
    @staticmethod
    def link_dir(link: Any, target: Any) -> None: ...
    @staticmethod
    def link_dirs(link_target_tuples: Any): ...
    @staticmethod
    def link_file(link: Any, target: Any) -> None: ...
    @staticmethod
    def link_files(link_target_tuples: Any): ...
    @staticmethod
    def unlink_dir(link: Any) -> None: ...
    @staticmethod
    def unlink_dirs(links: Any) -> None: ...
    @staticmethod
    def unlink_file(link: Any) -> None: ...
    @staticmethod
    def unlink_files(links: Any) -> None: ...
    sync: Any = ...

class DirTree:
    path: Any = ...
    parent: Any = ...
    is_last: Any = ...
    depth: Any = ...
    def __init__(self, path: str, parent_path: str, is_last: bool) -> Any: ...
    @classmethod
    def make_tree(
        cls,
        root: Any,
        parent: Optional[Any] = ...,
        is_last: bool = ...,
        criteria: Optional[Any] = ...,
    ): ...
    @property
    def displayname(self): ...
    def displayable(self): ...

pwd = getcwd
cd = chdir

def tree(dirpath: Any, criteria: Optional[Any] = ...): ...
def mv(src: Any, dst: Any) -> None: ...
def cp(
    src: Any, dst: Any, r: bool = ..., symlinks: bool = ..., ignore: Optional[Any] = ...
) -> None: ...
def cp_file(source: str, target: str) -> None: ...
def cp_dir(source: str, target: str) -> None: ...
def ls(dirpath: str = ..., abs: bool = ...) -> List[str]: ...
def ls_files(dirpath: str = ..., abs: bool = ...) -> List[str]: ...
def ls_dirs(dirpath: str = ..., abs: bool = ...) -> List[str]: ...
def ls_files_dirs(
    dirpath: str = ..., abs: bool = ...
) -> Tuple[List[str], List[str]]: ...
def rm(path_string: str, r: bool = ..., v: bool = ...) -> Any: ...
def basename(path_str: str) -> str: ...
def dirname(fdpath: str) -> str: ...
def export(key: str, val: Union[None, str] = ...) -> None: ...

path2name = basename
parent_dirpath = dirname
link_dir: Any
link_dirs: Any
link_file: Any
link_files: Any
unlink_dir: Any
unlink_dirs: Any
unlink_file: Any
unlink_files: Any
sync: Any
echo = print



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

_PKG_MAIN: str = """# -*- coding: utf-8 -*-
\"\"\"
=====================
Pretty Useful Package
=====================
\"\"\"

from .cli import main

if __name__ == "__main__":
    main()
    
"""
_PKG_INIT = """# -*- coding: utf-8 -*-
\"\"\"
=====================
Pretty Useful Package
=====================
\"\"\"
"""
_PKG_UTILS = """# -*- coding: utf-8 -*-
\"\"\"
===================
Pretty Useful Utils
===================
\"\"\"
# import sys
# from os import listdir
# from os import path

# from pupy import ljson      #$# load json from filepath (filepath)
# from pupy import lstr       #$# load str from filepath (filepath)
# from pupy import sjson      #$# save json to filepath (filepath, data)
# from pupy import sstr       #$# save str to filepath (filepath, string)
# from pupy import files_gen  #$# gen filepaths below directory (dirpath)
# from pupy import dirs_gen   #$# gen dirpaths below directory (dirpath)
# from pupy import sh         #$# shell-y funks; sh.cd, sh.ls, sh.pwd...

def main():
    pass

if __name__ == "__main__":
    main()
    # from doctest import testmod
    # testmod() 
"""
_IO_SCRIPT = """# -*- coding: utf-8 -*-
\"\"\"
====================
Pretty Useful Script
====================
\"\"\"
__version__ = "0.0.1"
import sys
from argparse import ArgumentParser
from argparse import FileType
from os import listdir
from os import path

# from pupy import ljson      #$# load json from filepath (filepath)
# from pupy import lstr       #$# load str from filepath (filepath)
# from pupy import sjson      #$# save json to filepath (filepath, data)
# from pupy import sstr       #$# save str to filepath (filepath, string)
# from pupy import files_gen  #$# gen filepaths below directory (dirpath)
# from pupy import dirs_gen   #$# gen dirpaths below directory (dirpath)
# from pupy import sh         #$# shell-y funks; sh.cd, sh.ls, sh.pwd...

PARSER = ArgumentParser(description="python scripty.py < stdin > stdout")

_INS = PARSER.add_mutually_exclusive_group()
_INS.add_argument(
    "-i",
    "--input",
    type=FileType("r"),
    default=sys.stdin,
    metavar="IN",
    help="Input file name; or stdin.",
)
_INS.add_argument(
    "strinput", nargs="?", type=str, metavar="STDIN", help="Input string via"
)
PARSER.add_argument(
    "-o",
    "--output",
    type=FileType("w"),
    help="Output file name (defaults to STDOUT)",
    default=sys.stdout,
)


def main():
    ARGV = PARSER.parse_args()
    input_str = ARGV.strinput or ARGV.input.read()
    ARGV.output.write(input_str)
    ARGV.output.write("OUTPUT")


if __name__ == "__main__":
    main()
    # from doctest import testmod
    # testmod() 
"""
_IO_PKG_CLI = _IO_SCRIPT.replace(
    "====================\n" "Pretty Useful Script\n" "====================\n",
    "=====================\n" "Pretty Useful Package\n" "=====================\n",
)

if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

from typing import Any
from typing import Optional

def base_cmd(args: Any) -> None: ...
def unescaped_str(arg_str: Any): ...
def new_package(relative_path: Any) -> None: ...
def new_cmd(args: Any) -> None: ...

PARSER: Any
SUBPARSERS: Any
NEW_SUBPARSER: Any

def main(ARGS: Optional[Any] = ...) -> None: ...



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python
"""
=========
Shell-ish
=========
"""
from distutils.dir_util import copy_tree
from glob import iglob
from os import chdir
from os import environ
from os import getcwd
from os import listdir
from os import lstat
from os import makedirs
from os import path
from os import readlink
from os import remove
from os import stat
from os import symlink
from os import unlink
from platform import system

from pathlib import Path
from shutil import copy2
from functools import lru_cache
from shutil import copystat
from shutil import move
from shutil import rmtree
from subprocess import PIPE
from subprocess import run
from typing import List
from typing import Tuple
from typing import Union

from pupy.foreign import chunks

class LIN:
    @staticmethod
    def rsync_args(
        src, dest, delete=False, mkdirs=False, exclude=[], include=[], dry_run=False
        ):
        """Sheldon rsync wrapper for syncing tdirs

        Args:
          dest: path to local tdir
          src: path to remote (raid) tdir

        Returns:
          subprocess return code from rsync

          Rsync return codes:

          - 0 == Success
          - 1 == Syntax or usage error
          - 2 == Protocol incompatibility
          - 3 == Errors selecting input/output files, dirs
          - 4 == Requested  action not supported: an attempt was made to manipulate
          64-bit files on a platform that cannot support them; or an option
          was specified that is supported by the client and not the server.
          - 5 == Error starting client-server protocol
          - 6 == Daemon unable to append to log-file
          - 10 == Error in socket I/O
          - 11 == Error in file I/O
          - 12 == Error in rsync protocol data stream
          - 13 == Errors with program diagnostics
          - 14 == Error in IPC code
          - 20 == Received SIGUSR1 or SIGINT
          - 21 == Some error returned by waitpid()
          - 22 == Error allocating core memory buffers
          - 23 == Partial transfer due to error
          - 24 == Partial transfer due to vanished source files
          - 25 == The --max-delete limit stopped deletions
          - 30 == Timeout in data send2viewserver/receive
          - 35 == Timeout waiting for daemon connection

        """
        if mkdirs:
            try:
                assert path.exists(dest) and path.isdir(dest)
            except AssertionError:
                makedirs(dest, exist_ok=True)

        if not dest.endswith("/"):
            dest = "{}/".format(dest)
        if not src.endswith("/"):
            src = "{}/".format(src)
        _args = [
            "rsync",
            "-a",
            "-O",
            "--no-o",
            "--no-g",
            "--no-p",
            "--delete" if delete else None,
            *('--exclude="{}"'.format(pattern) for pattern in exclude),
            *('--include="{}"'.format(pattern) for pattern in include),
            *(("--dry-run", "-i") if dry_run else (None,)),
            src,
            dest,
            ]
        return list(filter(None, _args))

    @staticmethod
    def rsync(src, dest, delete=False, exclude=[], include=[], dry_run=False):
        """Sheldon rsync wrapper for syncing tdirs

        :param delete:
        :param dest: path to local tdir
        :param src: path to remote (raid) tdir
        :return: subprocess return code from rsync

        Rsync return codes:

         - 0 == Success
         - 1 == Syntax or usage error
         - 2 == Protocol incompatibility
         - 3 == Errors selecting input/output files, dirs
         - 4 == Requested  action not supported: an attempt was made to manipulate
           64-bit files on a platform that cannot support them; or an option
           was specified that is supported by the client and not the server.
         - 5 == Error starting client-server protocol
         - 6 == Daemon unable to append to log-file
         - 10 == Error in socket I/O
         - 11 == Error in file I/O
         - 12 == Error in rsync protocol data stream
         - 13 == Errors with program diagnostics
         - 14 == Error in IPC code
         - 20 == Received SIGUSR1 or SIGINT
         - 21 == Some error returned by waitpid()
         - 22 == Error allocating core memory buffers
         - 23 == Partial transfer due to error
         - 24 == Partial transfer due to vanished source files
         - 25 == The --max-delete limit stopped deletions
         - 30 == Timeout in data send2viewserver/receive
         - 35 == Timeout waiting for daemon connection

        """
        rsync_args = LIN.rsync_args(
            src,
            dest,
            delete=delete,
            mkdirs=True,
            exclude=exclude,
            include=include,
            dry_run=dry_run,
            )
        subproc = run(args=list(filter(None, rsync_args)), stdout=PIPE, stderr=PIPE)
        return subproc

    @staticmethod
    def link_dir(linkpath, targetpath):
        symlink(targetpath, linkpath)

    @staticmethod
    def link_dirs(link_target_tuples):
        for link, target in link_target_tuples:
            LIN.link_dir(link, target)

    @staticmethod
    def link_file(linkpath: str, targetpath: str) -> None:
        makedirs(path.split(linkpath)[0], exist_ok=True)
        symlink(targetpath, linkpath)

    @staticmethod
    def link_files(link_target_tuples):
        for link, target in link_target_tuples:
            LIN.link_file(link, target)

    @staticmethod
    def unlink_dir(link):
        unlink(link)

    @staticmethod
    def unlink_dirs(links):
        for link in links:
            unlink(link)

    @staticmethod
    def unlink_file(link):
        unlink(link)

    @staticmethod
    def unlink_files(links):
        for link in links:
            unlink(link)

    sync = rsync

@lru_cache(maxsize=None)
def _check_link_target_files(link, target):
    # for link, target in link_target_tuples:
    try:
        assert path.exists(target)
        makedirs(path.split(link)[0], exist_ok=True)
        # _exists.extend(["mklink", link, target, "&&"])
        return True
        # _exists.extend(["mklink", link, target, "&&"])
    except AssertionError as e:
        print(
            "Link target not found; unable to create link:\n    {} => {}".format(
                link, target
                )
            )
    except Exception as e:
        print(e, type(e))
    return False

@lru_cache(maxsize=None)
def _check_link_target_dirs(link, target):
    # for link, target in link_target_tuples:
    try:
        assert path.exists(target) and path.isdir(target)
        # makedirs(path.split(link)[0], exist_ok=True)
        # _exists.extend(["mklink", link, target, "&&"])
        # _exists.extend(["mklink", link, target, "&&"])
    except AssertionError as e:
        print(
            "Link target not found; unable to create link:\n    {} => {}".format(
                link, target
                )
            )
        return False
    except Exception as e:
        print(e, type(e))
    try:
        rmtree(link)

    except:
        pass
    return True

class WIN:

    # def rsync_args(src, dest,
    #                delete=False, mkdirs=False,
    #                exclude=[], include=[], dry_run=False):
    @staticmethod
    def robocopy_args(
        src, dest, delete=False, exclude_files=[], exclude_dirs=[], dry_run=False
        ):
        """Robocopy for sheldon

        Args:
          dest: path to local tdir
          src: path to remote (raid) tdir

        Returns:
          subprocess return code from robocopy

          Robocopy return codes:

          0. No files were copied. No failure was encountered. No files were
          mismatched. The files already exist in the destination directory;
          therefore, the copy operation was skipped.
          1. All files were copied successfully.
          2. There are some additional files in the destination directory that are not
          present in the source directory. No files were copied.
          3. Some files were copied. Additional files were present. No failure
          was encountered.
          5. Some files were copied. Some files were mismatched. No failure was
          encountered.
          6. Additional files and mismatched files exist. No files were copied
          and no failures were encountered. This means that the files already
          exist in the destination directory.
          7. Files were copied, a file mismatch was present, and additional files
          were present.
          8. Several files did not copy.

        """
        _args = [
            "robocopy",
            src,
            dest,
            "/MIR" if delete else "/E",
            "/mt",
            "/W:1",
            "/R:1",
            ]
        if exclude_dirs:
            _args.extend(["/XD", *exclude_dirs])
        if exclude_files:
            _args.extend(["/XF", *exclude_files])
        if dry_run:
            _args.append("/L")
        return list(filter(None, _args))

    @staticmethod
    def robocopy(
        src, dest, delete=False, exclude_files=[], exclude_dirs=[], dry_run=False
        ):
        """Robocopy for sheldon

        :param dest: path to local tdir
        :param src: path to remote (raid) tdir
        :return: subprocess return code from robocopy

        Robocopy return codes:

        0. No files were copied. No failure was encountered. No files were
           mismatched. The files already exist in the destination directory;
           therefore, the copy operation was skipped.
        1. All files were copied successfully.
        2. There are some additional files in the destination directory that are not
           present in the source directory. No files were copied.
        3. Some files were copied. Additional files were present. No failure
           was encountered.
        5. Some files were copied. Some files were mismatched. No failure was
           encountered.
        6. Additional files and mismatched files exist. No files were copied
           and no failures were encountered. This means that the files already
           exist in the destination directory.
        7. Files were copied, a file mismatch was present, and additional files
           were present.
        8. Several files did not copy.


        """
        _args = WIN.robocopy_args(
            src=src,
            dest=dest,
            delete=delete,
            exclude_files=exclude_files,
            exclude_dirs=exclude_dirs,
            dry_run=dry_run,
            )
        subproc = run(args=_args, stdout=PIPE, stderr=PIPE)
        return subproc

    @staticmethod
    def link_dir(link, target):
        makedirs(link, exist_ok=True)
        run(args=["mklink", "/D", link, target], stdout=PIPE, stderr=PIPE, shell=True)

    @staticmethod
    def link_dirs(link_target_tuples):
        _exists = [
            "mklink /D {} {}".format(link, target)
            for link, target in link_target_tuples
            if _check_link_target_dirs(link, target)
            ]
        _proc = run(args=" && ".join(_exists).split(" "), stdout=PIPE, stderr=PIPE, shell=True)
        stderr = _proc.stderr.decode()
        if 'too' in stderr and 'long' in stderr:
            tuple_chunks = list(chunks(link_target_tuples, len(link_target_tuples) // 2))
            for tuple_chunk in tuple_chunks:
                WIN.link_dirs(tuple_chunk)

    @staticmethod
    def link_file(link, target):
        makedirs(path.split(link)[0], exist_ok=True)
        run(args=["mklink", link, target], stdout=PIPE, stderr=PIPE, shell=True)

    @staticmethod
    def link_files(link_target_tuples):
        link_target_tuples = list(link_target_tuples)
        _exists = [
            "mklink {} {}".format(link, target)
            for link, target in link_target_tuples
            if _check_link_target_files(link, target)
            ]
        _proc = run(args=" && ".join(_exists).split(' '), stdout=PIPE, stderr=PIPE, shell=True)
        stdout = _proc.stdout.decode()
        stderr = _proc.stderr.decode()
        if 'too' in stderr and 'long' in stderr:
            tuple_chunks = list(chunks(link_target_tuples, len(link_target_tuples) // 2))
            for tuple_chunk in tuple_chunks:
                WIN.link_files(tuple_chunk)

    @staticmethod
    def unlink_dir(link):
        run(args=["RD", link], stdout=PIPE, stderr=PIPE, shell=True)

    @staticmethod
    def unlink_dirs(links):
        cmd_args = " && ".join("RD {}".format(link) for link in links).split(" ")
        run(args=cmd_args, stdout=PIPE, stderr=PIPE, shell=True)

    @staticmethod
    def unlink_file(link):
        run(args=["Del", link], stdout=PIPE, stderr=PIPE, shell=True)

    @staticmethod
    def unlink_files(links):
        cmd_args = " && ".join("Del {}".format(link) for link in links).split(" ")
        run(args=cmd_args, stdout=PIPE, stderr=PIPE, shell=True)

    sync = robocopy

class DirTree:
    _filename_prefix_mid: str = "├──"
    _filename_prefix_last: str = "└──"
    _parent_prefix_middle: str = "    "
    _parent_refix_last: str = "│   "

    def __init__(self, path: str, parent_path: str, is_last: bool):
        self.path = Path(str(path))
        self.parent = parent_path
        self.is_last = is_last
        self.depth: int = self.parent.depth + 1 if self.parent else 0

    @classmethod
    def make_tree(cls, root, parent=None, is_last=False, criteria=None):
        root = Path(str(root))
        criteria = criteria or cls._default_criteria

        displayable_root = cls(root, parent, is_last)
        yield displayable_root

        children = sorted(
            list(path for path in root.iterdir() if criteria(str(path))),
            key=lambda s: str(s).lower(),
            )
        count = 1
        for path in children:
            is_last = count == len(children)
            if path.is_dir():
                yield from cls.make_tree(
                    path, parent=displayable_root, is_last=is_last, criteria=criteria
                    )
            else:
                # print(path)
                yield cls(path, displayable_root, is_last)
            count += 1

    @classmethod
    def _default_criteria(cls, path_string):
        ignore_strings = (".pyc", "__pycache__")
        return not any(
            ignored in str(path_string).lower() for ignored in ignore_strings
            )

    @property
    def displayname(self):
        if self.path.is_dir():
            return self.path.name + "/"
        return self.path.name

    def displayable(self):
        if self.parent is None:
            return self.displayname

        _filename_prefix = (
            self._filename_prefix_last if self.is_last else self._filename_prefix_mid
        )

        parts = ["{!s} {!s}".format(_filename_prefix, self.displayname)]

        parent = self.parent
        while parent and parent.parent is not None:
            parts.append(
                self._parent_prefix_middle
                if parent.is_last
                else self._parent_refix_last
                )
            parent = parent.parent

        return "".join(reversed(parts))

# _OS = system().lower()
_OS = WIN if "windows" in system().lower() else LIN

pwd = getcwd
cd = chdir

def tree(dirpath, criteria=None):
    return "\n".join(
        p.displayable() for p in DirTree.make_tree(Path(dirpath), criteria=criteria)
        )

def mv(src, dst):
    for file in iglob(src, recursive=True):
        move(file, dst)

def cp(src, dst, r=False, symlinks=False, ignore=None):
    makedirs(dst, exist_ok=True)
    if not path.exists(dst):
        makedirs(dst, exist_ok=True)
        copystat(src, dst)
    _listdir = listdir(src)
    if ignore:
        _listdir = [x for x in _listdir if x not in set(ignore(src, _listdir))]
    for item in _listdir:
        _src_pth = path.join(src, item)
        _dest_pth = path.join(dst, item)
        if symlinks and path.islink(_src_pth):
            if path.lexists(_dest_pth):
                remove(_dest_pth)
            symlink(readlink(_src_pth), _dest_pth)
            try:
                from os import lchmod

                st = lstat(_src_pth)
                mode = stat.S_IMODE(st.st_mode)
                lchmod(_dest_pth, mode)
            except:
                pass
        elif path.isdir(_src_pth):
            if r:
                cp(_src_pth, _dest_pth, r=True, symlinks=symlinks, ignore=ignore)
            else:
                print("{} is dir; use rm(..., r=True)".format(_src_pth))
        else:
            copy2(_src_pth, _dest_pth)

def cp_file(source: str, target: str) -> None:
    dirname = path.dirname(target)
    makedirs(dirname, exist_ok=True)
    copy2(source, target)

def cp_dir(source: str, target: str) -> None:
    if not path.exists(target):
        makedirs(target)
    copy_tree(source, target)

def _cp_dest(source: str, target: str) -> str:
    if path.isfile(source) and path.isdir(target):
        return path.join(target, path.basename(source))
    return target

def cp(source: str, target: str, f: bool = True, r: bool = False) -> None:
    """Copy the directory/file src to the directory/file target"""
    if (path.exists(target) and not f) or source == target:
        return
    if path.isdir(source) and not r:
        raise ValueError("Source ({}) is directory; use r=True")
    target = _cp_dest(source, target)
    if path.isfile(source) or path.islink(source):
        cp_file(source, target)
    if path.isdir(source):
        cp_dir(source, target)

def ls(dirpath: str = ".", abs: bool = False) -> List[str]:
    if abs:
        return [path.join(dirpath, item) for item in listdir(dirpath)]
    return listdir(dirpath)

def ls_files(dirpath: str = ".", abs: bool = False) -> List[str]:
    files = (file for file in ls(dirpath, abs=True) if path.isfile(file))
    if not abs:
        return list(map(lambda el: el.replace(dirpath, "."), files))
    return list(files)

def ls_dirs(dirpath: str = ".", abs: bool = False) -> List[str]:
    dirs = (dir for dir in ls(dirpath, abs=True) if path.isdir(dir))
    if not abs:
        return list(map(lambda el: el.replace(dirpath, "."), dirs))
    return list(dirs)

def ls_files_dirs(dirpath: str = ".", abs: bool = False) -> Tuple[List[str], List[str]]:
    return ls_files(dirpath, abs=abs), ls_dirs(dirpath, abs=abs)

def rm(path_string: str, r: bool = False, v: bool = False):
    """rm should act like the (ba)sh-rm

    This function was implemented by my cousin Matty-Ice (AKA Matt Bommer)

    :param f_arg:
    :type f_arg:
    :param args:
    :type args:
    :return:
    :rtype:

    """
    for _path_str in iglob(path_string, recursive=True):
        try:
            remove(_path_str)
            if v:
                print("Removed file: {}".format(_path_str))

        except Exception as e:
            if r:
                rmtree(_path_str)
                if v:
                    print("Removed dir: {}".format(_path_str))
            else:
                raise ValueError(_path_str + " is a directory -- use r=True")

def basename(path_str: str) -> str:
    """Get the parent-directory for a file or directory path as a string

    :param path_str:
    :return: The parent directory as a string

    .. doctest:: python

        >>> from os import getcwd
        >>> basename(getcwd()) in getcwd()
        True

    """
    return path.split(path.abspath(path_str))[-1]

def dirname(fdpath: str) -> str:
    """Return the parent directory for the given file or dir path

    :param fdpath: file/dir-path as as string
    :return: parent directory for the given file or dir path


    .. doctest:: python

        >>> from os import path
        >>> dirname(path.abspath(__file__)) in path.abspath(__file__)
        True

    """
    return path.split(fdpath)[0]

def export(key: str, val: Union[None, str] = None) -> None:
    if val:
        environ[key] = val
        return
    if "=" in key:
        _key, *val = key.split("=")
        export(_key, "=".join(val))

path2name = basename  # Alias for basename
parent_dirpath = dirname  # Alias for dirname

# Operating system dependent things
link_dir = _OS.link_dir
link_dirs = _OS.link_dirs
link_file = _OS.link_file
link_files = _OS.link_files
unlink_dir = _OS.unlink_dir
unlink_dirs = _OS.unlink_dirs
unlink_file = _OS.unlink_file
unlink_files = _OS.unlink_files
sync = _OS.sync
echo = print



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python
"""
==========
Iter Funks
==========

foreign => 'for ... in ...' (its a pun)
"""
from collections import Counter
from collections import deque
from functools import reduce
from itertools import chain
from itertools import tee
from operator import mul
from os import getcwd
from os import path
from os import sep
from os import walk
from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Union

from pupy._typing import Flint
from pupy._typing import Paths


def files_gen(dirpath: str = getcwd(), abspath: bool = True) -> Paths:
    """Yields paths beneath dirpath param; dirpath defaults to os.getcwd()

    :param dirpath: Directory path to walking down/through.
    :param abspath: Yield the absolute path
    :return: Generator object that yields filepaths (absolute or relative)

    """
    return (
        fpath if abspath else fpath.replace(dirpath, "").strip(sep)
        for fpath in (
            path.join(pwd, file) for pwd, dirs, files in walk(dirpath) for file in files
        )
    )


def dirs_gen(dirpath: str = getcwd(), abspath: bool = True) -> Paths:
    """Yields paths beneath dirpath param; dirpath defaults to os.getcwd()

    :param dirpath: Directory path to walking down/through.
    :param abspath: Yield the absolute path
    :return: Generator object that yields dir-paths (absolute or relative)

    """
    return (
        dirpath if abspath else dirpath.replace(dirpath, "").strip(sep)
        for dirpath in (pwd for pwd, dirs, files in walk(dirpath))
    )


def files_dirs_gen(
    dirpath: str = getcwd(), abspath: bool = True
) -> Tuple[Paths, Paths]:
    return (files_gen(dirpath, abspath=abspath), dirs_gen(dirpath, abspath=abspath))


def walk_gen(dirpath: str = getcwd(), abspath: bool = True) -> Paths:
    """Yields all paths beneath dirpath param; dirpath defaults to os.getcwd()

    :param dirpath: Directory path to walking down/through.
    :param abspath: Yield the absolute path
    :return: Generator object that yields filepaths (absolute or relative)

    """
    return chain.from_iterable(
        fpath if abspath else fpath.replace(dirpath, "").strip(sep)
        for fpath in (
            chain.from_iterable(
                (
                    (pwd,),
                    (path.join(pwd, _dir) for _dir in dirs),
                    (path.join(pwd, _file) for _file in files),
                )
                for pwd, dirs, files in walk(dirpath)
            )
        )
    )


def exhaust(it: Iterable[Any]) -> None:
    """Exhaust an interable / use it up; useful for evaluating a map object.

    :param it: iterable object to run through
    :return: None

    .. docstring::python

        >>> a = [1, 2, 3, 4, 5, 6]
        >>> a_map = map(lambda x: x*2, a)
        >>> a_exhausted = exhaust(a_map)
        >>> a_exhausted == None # will be none after being exhausted
        True

    """
    deque(it, maxlen=0)


def chunks(it: Union[List[int], str], chunk_size: int) -> Iterable[Any]:
    """Yields chunks of something slicable with length <= chunk_size

    :param it:
    :param chunk_size: size of the chunks
    :type chunk_size: int

    .. docstring::python

        >>> list(chunks([1, 2, 3, 4, 5, 6], 3))
        [[1, 2, 3], [4, 5, 6]]
        >>> list(chunks([1, 2, 3, 4, 5, 6], 2))
        [[1, 2], [3, 4], [5, 6]]
        >>> list(chunks('abcdefghijklmnopqrstuvwxyz', 2))
        ['ab', 'cd', 'ef', 'gh', 'ij', 'kl', 'mn', 'op', 'qr', 'st', 'uv', 'wx', 'yz']
        >>> list(chunks('abcdefghijklmnopqrstuvwxyz', 13))
        ['abcdefghijklm', 'nopqrstuvwxyz']

    """
    return (it[i : i + chunk_size] for i in range(0, len(it), chunk_size))


def is_permutation(
    a: Union[int, List[int], str], b: Union[int, List[int], str]
) -> bool:
    """Checks if two integers or lists are permutations lists are permutations

    :param a: possible perumtation of b
    :type a: int or list
    :param b: possible perumtation of a
    :type b: int or list
    :returns: True if a and b are permutations of one another; False otherwise
    :rtype: bool

    It works for list!!!

    .. doctest:: python

        >>> a = [1, 2, 3, 4]
        >>> b = [4, 3, 2, 1]
        >>> is_permutation(a, b)
        True
        >>> a = [1, 3, 2, 4]
        >>> b = [4, 3, 2, 1]
        >>> is_permutation(a, b)
        True
        >>> a = [1, 3, 2, 4]
        >>> b = [4, 3, 2, 1, 1]
        >>> is_permutation(a, b) # False cuz of the extra 'one'
        False

    It works for integers!?!?!?

    .. doctest:: python

        >>> a = 1234
        >>> b = 4321
        >>> is_permutation(a, b)
        True
        >>> a = 12344
        >>> b = 43212
        >>> is_permutation(a, b)
        False

    It also works for strings!!!

    .. doctest:: python

        >>> a = 'abcd'
        >>> b = 'dbca'
        >>> is_permutation(a, b) # False cuz of the extra 'one'
        True
        >>> a = 'pood'
        >>> b = 'doop'
        >>> is_permutation(a, b) # False cuz of the extra 'one'
        True
        >>> a = 'snorkel'
        >>> b = 'doop'
        >>> is_permutation(a, b) # False cuz of the extra 'one'
        False

    """

    if isinstance(a, int):
        a = digits_list(a)
    if isinstance(b, int):
        b = digits_list(b)
    return len(a) == len(b) and Counter(a) == Counter(b)


def rotate(rlist: List[int], rn: int = 1, left_rotate: bool = True) -> List[int]:
    """Rotate a list (rlist) by rn indices to the left or right

    :param rlist: list/toople or slicable to be rotated
    :type rlist: list or tuple
    :param rn: steps bywhich to rotate (Default value = 1)
    :type rn: int
    :param left_rotate: True (default) left rotates; False right rotates.
    :type left_rotate: bool
    :returns: rotated list
    :rtype: list

    .. doctest:: python

        >>> rotate([1, 2, 3, 4], left_rotate=True)
        [2, 3, 4, 1]
        >>> rotate([1, 2, 3, 4], left_rotate=False)
        [4, 1, 2, 3]
        >>> rotate([1, 2, 3, 4], rn=4, left_rotate=False)
        [1, 2, 3, 4]

    """

    def _left_rotate(l, n=1):
        """

        :param l:
        :param n:  (Default value = 1)

        """
        return l[n:] + l[:n]

    def _right_rotate(l, n=1):
        """

        :param l:
        :param n:  (Default value = 1)

        """
        return l[-n:] + l[:-n]

    return _left_rotate(rlist, rn) if left_rotate else _right_rotate(rlist, rn)


def rotations_gen(rlist: Iterable[Any]) -> Iterable[Any]:
    """Yields all rotations of a list

    :param rlist:

    .. doctest::python

        >>> for rot in rotations_gen((1, 2, 3, 4)):
        ...     print(rot)
        ...
        (1, 2, 3, 4)
        (4, 1, 2, 3)
        (3, 4, 1, 2)
        (2, 3, 4, 1)

    """
    return ((rlist[-i:] + rlist[:-i]) for i in range(len(rlist)))


def digits_list(number: int) -> List[int]:
    """Returns a list of the digits in num

    :param number: number w/ digits to be listsed
    :type number: int
    :returns: -> digits in a list
    :rtype: list

    .. doctest::python

        >>> digits_list(1111)
        [1, 1, 1, 1]
        >>> digits_list(982)
        [9, 8, 2]
        >>> digits_list(101)
        [1, 0, 1]
        >>> digits_list(123)
        [1, 2, 3]

    """

    digits = deque()
    for _ in range(len(str(number))):
        number, r = divmod(number, 10)
        digits.appendleft(r)
    return list(digits)


def int_from_digits(digits: Iterable[int]) -> int:
    """Converts an iterable of digits digits to a number

    The iteratble can be ints or strings/chars

    :rtype: int

    .. doctest::python

        >>> int_from_digits([3, 2, 1])
        321
        >>> int_from_digits([1, 1, 1, 1, 2, 3])
        111123
        >>> int_from_digits([1, 2, 3])
        123

    """
    return sum(
        digits[len(list(digits)) - i - 1] * 10 ** i
        for i in range(0, len(list(digits)), 1)
    )


def iter_product(l: Iterable[int]) -> Flint:
    """Product of all the elements in a list or tuple

    :param l: list with integer elements
    :returns: product of all the elements in a list
    :rtype: int

    .. doctest::python

        >>> iter_product([1, 2, 3, 4])
        24
        >>> iter_product(tuple([1, 2, 3, 4]))
        24
        >>> iter_product([-1, -2, -3, 4])
        -24

    """
    return reduce(mul, l)


def spliterable(
    iterable: Iterable[Any], funk: Callable[[Any], bool]
) -> Tuple[Iterable[Any], Iterable[Any]]:
    """

    :param iterable:
    :type iterable:
    :param funk:
    :type funk:
    :return:
    :rtype:

    .. doctest::python

        >>> is_even = lambda n: n % 2 == 0
        >>> a, b = spliterable(range(10), is_even)
        >>> list(a)
        [0, 2, 4, 6, 8]
        >>> list(b)
        [1, 3, 5, 7, 9]

    """
    _true_gen, _false_gen = tee((funk(item), item) for item in iterable)
    return (i for p, i in _true_gen if p), (i for p, i in _false_gen if not p)


if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python
from bisect import bisect
from bisect import bisect_right
from collections.abc import MutableSequence
from itertools import count
from math import sqrt
from typing import Any
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Union

from pupy.decorations import cash_it
from pupy.maths import divisors_gen


def prime_gen(
    plim: int = 0, kprimes: Union[None, Iterable[int]] = None
) -> Iterator[int]:
    """Infinite (within reason) prime number generator
    
    My big modification is the pdiv_dictionary() function that recreats the
    dictionary of divisors so that you can continue to generate prime numbers
    from a (sorted) list of prime numbers.
    
    Based on:
        eratosthenes by David Eppstein, UC Irvine, 28 Feb 2002
        http://code.activestate.com/recipes/117119/ and the thread at the url

    :param plim: prime_limit; default=0 makes for an infinite generator
    :type plim: int
    :param kprimes: known_primes as an iterable (Default value = None)
    :type kprimes: iter

    .. doctest::

        >>> list(prime_gen(50))
        [2, 3, 5, 7, 11, 13, 17, 19, 23, 29, 31, 37, 41, 43, 47]
        >>> list(prime_gen(10))
        [2, 3, 5, 7]

    """

    if kprimes is None:
        kprimes = [2, 3, 5, 7, 11]

    def _dictionary():
        """Recreates the prime divisors dictionary used by the generator"""
        div_dict = {}
        for pdiv in kprimes:
            multiple = kprimes[-1] // pdiv * pdiv
            if multiple % 2 == 0:
                multiple += pdiv
            else:
                multiple += 2 * pdiv
            while multiple in div_dict:
                multiple += pdiv * 2
            div_dict[multiple] = pdiv
        return div_dict

    # [1]
    # See if the upper bound is greater than the known primes
    if 0 < plim <= kprimes[-1]:
        for p in kprimes:
            if p <= plim:
                yield p
        return  # return bc we are done

    # [2]
    # Recreate the prime divisibility dictionary using kprimes;
    # Set start and yield first 4 primes
    divz = _dictionary()
    start = kprimes[-1] + 2  # max prime + 2 (make sure it is odd)
    if start == 13:
        yield 2
        yield 3
        yield 5
        yield 7
        yield 11
    # use count or range depending on if generator is infinite
    it = count(start, 2) if plim == 0 else range(start, plim, 2)
    for num in it:
        prime_div = divz.pop(num, None)
        if prime_div:
            multiple = (2 * prime_div) + num
            while multiple in divz:
                multiple += 2 * prime_div
            divz[multiple] = prime_div
        else:
            divz[num * num] = num
            yield num


def prime_factorization_gen(n: int) -> Iterator[int]:
    """generates all numbers in the prime factorization of n

    :param n: number to be factored
    :type n: int

    .. doctest::

        >>> list(prime_factorization_gen(12))
        [2, 2, 3]
        >>> list(prime_factorization_gen(16))
        [2, 2, 2, 2]

    """
    for factor in prime_factors_gen(n):
        if n <= 1:
            break
        while n % factor == 0:
            n //= factor
            yield factor


def prime_factors_gen(n: int) -> Iterator[Any]:
    """prime factors generator

    :param n: number to be factorized
    :type n: int

    .. doctest:: python

        >>> list(prime_factors_gen(12))
        [2, 3]
        >>> list(prime_factors_gen(16))
        [2]

    """
    return (p for p in divisors_gen(n) if is_prime(p))


@cash_it
def is_prime(number: int) -> bool:
    """Checks if a number is prime given that number as a param

    :param number: number to check if is prime
    :type number: int
    :returns: -> True if number is prime
    :rtype: bool
    


        >>> is_prime(1)
        False
        >>> is_prime(2)
        True
        >>> is_prime(3)
        True
        >>> is_prime(4)
        False
        >>> is_prime(5)
        True
        >>> is_prime(6)
        False
        >>> is_prime(7)
        True
        >>> is_prime(100)
        False
        >>> is_prime(89)
        True

    """
    if number == 2 or number == 3:
        return True
    if number < 2 or number % 2 == 0:
        return False
    if number < 9:
        return True
    if number % 3 == 0:
        return False
    for step in range(5, int(sqrt(number)) + 1, 6):
        if step >= number:
            break
        if number % step == 0:
            return False
        if number % (step + 2) == 0:
            return False
    return True


class OctopusPrime(MutableSequence):
    """OctopusPrime, the 8-leg autobot, here to help you find PRIMES

    ..
    
        ───────────▄▄▄▄▄▄▄▄▄───────────
        ────────▄█████████████▄────────
        █████──█████████████████──█████
        ▐████▌─▀███▄───────▄███▀─▐████▌
        ─█████▄──▀███▄───▄███▀──▄█████─
        ─▐██▀███▄──▀███▄███▀──▄███▀██▌─
        ──███▄▀███▄──▀███▀──▄███▀▄███──
        ──▐█▄▀█▄▀███─▄─▀─▄─███▀▄█▀▄█▌──
        ───███▄▀█▄██─██▄██─██▄█▀▄███───
        ────▀███▄▀██─█████─██▀▄███▀────
        ───█▄─▀█████─█████─█████▀─▄█───
        ───███────────███────────███───
        ───███▄────▄█─███─█▄────▄███───
        ───█████─▄███─███─███▄─█████───
        ───█████─████─███─████─█████───
        ───█████─████─███─████─█████───
        ───█████─████─███─████─█████───
        ───█████─████▄▄▄▄▄████─█████───
        ────▀███─█████████████─███▀────
        ──────▀█─███─▄▄▄▄▄─███─█▀──────
        ─────────▀█▌▐█████▌▐█▀─────────
        ────────────███████────────────

    
    
    
    """

    def __init__(self, plim: int = 100) -> None:
        # list.__init__(self, list(prime_gen(plim=plim)))
        # super(OctopusPrime, self).__init__()
        p = [
            2,
            3,
            5,
            7,
            11,
            13,
            17,
            19,
            23,
            29,
            31,
            37,
            41,
            43,
            47,
            53,
            59,
            61,
            67,
            71,
            73,
            79,
            83,
            89,
            97,
        ]
        if plim == 100:
            self._list = p[:]
        elif plim < 100:
            self._list = list(filter(lambda n: n <= plim, p))
        else:
            self._list = list(prime_gen(plim, p))
        self.max_loaded = self._list[-1]

    def _transform(self, n: Optional[int] = None) -> None:
        """TRANSFORM / grow the list

        :param n:  (Default value = None)

        """
        n = n if n is not None else self._list[-1] * 10
        self._list.extend(list(prime_gen(plim=n, kprimes=self._list)))

    def primes_below(self, upper_bound):
        """Lists primes, p, such that  p < upper_bound

        :param upper_bound: exclusive upper bound
        :type upper_bound: int
        :returns: -> primes less than upper_bound
        :rtype: list

        """
        return self.primes_between(1, upper_bound)

    def primes_between(self, lower_bound: int, upper_bound: int) -> List[int]:
        """Lists primes, p, such that, lower_bound < p < upper_bound

        :param lower_bound: exclusive lower bound
        :type lower_bound: int
        :param upper_bound: exclusive upper bound
        :type upper_bound: int
        :returns: -> primes between lower_bound and upper_bound
        :rtype: list

        """
        if upper_bound > self[-1]:
            self._transform(upper_bound)
        return self[bisect_right(self, lower_bound) : bisect(self, upper_bound)]

    def __len__(self) -> int:
        return len(self._list)

    def __getitem__(self, i: Union[int, slice]) -> Union[int, List[int]]:
        return self._list[i]

    def __delitem__(self, i):
        del self._list[i]

    def __setitem__(self, key, value):
        self._list[key] = value

    def insert(self, index, object):
        """

        :param index: 
        :param object: 

        """
        self._list.insert(index)

    def __str__(self):
        return str(self._list)

    def __repr__(self):
        return str(self._list)


if __name__ == "__main__":
    from doctest import testmod

    testmod()



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import Tuple
from typing import Union

from pupy._typing import Flint
from pupy._typing import Paths

def files_gen(dirpath: str = ..., abspath: bool = ...) -> Paths: ...
def dirs_gen(dirpath: str = ..., abspath: bool = ...) -> Paths: ...
def files_dirs_gen(dirpath: str = ..., abspath: bool = ...) -> Tuple[Paths, Paths]: ...
def walk_gen(dirpath: str = ..., abspath: bool = ...) -> Paths: ...
def exhaust(it: Iterable[Any]) -> None: ...
def chunks(it: Union[List[int], str], chunk_size: int) -> Iterable[Any]: ...
def is_permutation(
    a: Union[int, List[int], str], b: Union[int, List[int], str]
) -> bool: ...
def rotate(rlist: List[int], rn: int = ..., left_rotate: bool = ...) -> List[int]: ...
def rotations_gen(rlist: Iterable[Any]) -> Iterable[Any]: ...
def digits_list(number: int) -> List[int]: ...
def int_from_digits(digits: Iterable[int]) -> int: ...
def iter_product(l: Iterable[int]) -> Flint: ...
def spliterable(
    iterable: Iterable[Any], funk: Callable[[Any], bool]
) -> Tuple[Iterable[Any], Iterable[Any]]: ...



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python
from contextlib import contextmanager
from datetime import datetime
from inspect import stack
from os import environ
from os import makedirs
from os import path
from shutil import rmtree
from tempfile import mkdtemp
from time import time
from typing import Any
from typing import Optional

from pupy._alias import pp
from pupy.sh import cd
from pupy.sh import link_dirs
from pupy.sh import link_files
from pupy.sh import unlink_dirs
from pupy.sh import unlink_files


def timestamp(ts: Optional[float] = None) -> str:
    """Time stamp string w/ format yyyymmdd-HHMMSS

    :return: timestamp string

    .. doctest:: python

        >>> from datetime import datetime
        >>> stamps = ['20190225-161151', '20190225-081151']
        >>> timestamp(1551111111.111111) in stamps
        True
        >>> datetime.now().strftime("%Y%m%d-%H%M%S") == timestamp()
        True
        >>> timestamp(datetime.now()) == timestamp()
        True

    """
    if ts is None:
        return datetime.now().strftime("%Y%m%d-%H%M%S")
    elif isinstance(ts, float):
        return datetime.fromtimestamp(ts).strftime("%Y%m%d-%H%M%S")
    elif isinstance(ts, datetime):
        return ts.strftime("%Y%m%d-%H%M%S")


def environ_dict():
    return {k: environ[k] for k in environ}


@contextmanager
def linked_tmp_dir(
    suffix=None, prefix=None, dir=None, mkdirs=[], lndirs=[], lnfiles=[]
):
    temp_dir = mkdtemp(suffix, prefix, dir)
    lnfiles = [
        (path.join(temp_dir, _rel_link), target) for _rel_link, target in lnfiles
    ]
    lndirs = [(path.join(temp_dir, _rel_link), target) for _rel_link, target in lndirs]
    # print(mkdirs)
    _dirs2make = [
        path.join(temp_dir, e)
        for e in (
            dirpath if isinstance(dirpath, str) else path.join(*dirpath)
            for dirpath in mkdirs
        )
    ]
    _dirs2make.extend((path.split(link)[0] for link, target in lnfiles))
    _dirs2make.extend((path.split(link)[0] for link, target in lndirs))
    for dirpath_route in _dirs2make:
        # print("mkingdir", dirpath_route)
        makedirs(path.join(temp_dir, dirpath_route), exist_ok=True)

    link_files(lnfiles)
    link_dirs(lndirs)
    # from pupy.foreign import files_gen, dirs_gen
    # from pprint import pprint
    # pprint(list(files_gen(temp_dir)))
    # pprint(list(dirs_gen(temp_dir)))
    # try:
    #     lndirs = (
    #         (path.join(temp_dir, _rel_link), target) for _rel_link, target in lndirs
    #     )
    # except TypeError as e:
    #     pass
    try:
        yield temp_dir
    finally:
        try:
            unlink_files(lnfiles)
        except Exception as e:
            pass
        try:
            unlink_dirs(lndirs)
        except Exception as e:
            pass
        try:
            rmtree(temp_dir)
        except PermissionError:
            # sleep(3)
            # print(pwd())
            # print(temp_dir)
            cd("..")
            # print(pwd())
            rmtree(temp_dir)


def prinfo(obj: Any) -> None:
    try:
        pp({"object": obj, "type": obj})
    except:
        print("object:\n{}".format(obj))
        print("type:\n{}".format(type(obj)))


def pyfilepath(split=False):
    _filepath = path.abspath(stack()[1][1])
    if split:
        return path.split(_filepath)
    return _filepath


def time_funk(funk, *args, **kwargs):
    ti = time()
    _ret = funk(*args, **kwargs)
    tf = time()
    return _ret, tf - ti


def cmp_funks(f1, f2, runs=1, *args, **kwargs):
    f1_time = 0
    f2_time = 0
    for i in range(runs):
        r1, f1t = time_funk(f1, *args, **kwargs)
        f1_time += f1t
        r2, f2t = time_funk(f2, *args, **kwargs)
        f2_time += f2t
    f1_time = f1_time / runs
    f2_time = f2_time / runs
    return {
        "f1": str(f1.__name__),
        "f2": str(f2.__name__),
        "f1-time": f1_time,
        "f2-time": f2_time,
        "f1/f2": f1_time / f2_time,
    }


def time_funk(funk, *args, **kwargs):
    ti = time()
    _ret = funk(*args, **kwargs)
    tf = time()
    return _ret, tf - ti


def cmp_funks(f1, f2, runs, *args, **kwargs):
    f1_time = 0
    f2_time = 0
    for i in range(runs):
        r1, f1t = time_funk(f1, *args, **kwargs)
        f1_time += f1t
        r2, f2t = time_funk(f2, *args, **kwargs)
        f2_time += f2t
    f1_time = f1_time / runs
    f2_time = f2_time / runs
    return {
        "f1": str(f1.__name__),
        "f2": str(f2.__name__),
        "f1-time": f1_time,
        "f2-time": f2_time,
        "f1/f2": f1_time / f2_time,
        "runs": runs,
    }



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python


def _json():
    for mod in ("rapidjson", "ujson", "json"):
        try:
            res = __import__(mod)
            return res
        except ImportError:
            pass
    raise ImportError(
        "Ya need one of the following:\n"
        "   ~ ujson; 'pip install ujson'\n"
        "   ~ ujson; 'pip install python-rapidjson'\n"
        "   ~ json; this should have come with python\n"
    )


json = _json()



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python
from pathlib import Path

__version__ = (
    [
        l
        for l in open(str(Path(__file__).resolve().parents[1] / "pyproject.toml"))
        .read()
        .split("\n")
        if "version" in l
    ][0]
    .replace("version = ", "")
    .strip('"')
)



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python

from typing import Any

from pupy._typing import JASM

def asyncify(funk: Any): ...
async def sabytes(filepath: str, _bytes: bytes) -> None: ...
async def labytes(filepath: str) -> bytes: ...
async def sastring(filepath: str, string: str) -> None: ...
async def lastring(filepath: str) -> str: ...
async def sajson(filepath: str, data: JASM, minify: bool = ...) -> None: ...
async def lajson(filepath: str) -> JASM: ...
async def latoml(filepath: str) -> JASM: ...
async def satoml(filepath: str, data: JASM) -> None: ...



# -*- coding: utf-8 -*-
# Pretty ~ Useful ~ Python
from sys import stdout


def yesno(question, default=True, tries=3):
    """Ask a yes/no question and return an answer as a boolean

    :param question: question to ask a user
    :param default: True=yes; False=no; None=no-default
    :param tries: number of tries before giving up
    :return: True/False depending on the response

    """
    valid = {"yes": True, "y": True, "ye": True, "no": False, "n": False}
    default_prompts = {None: "[y/n]", True: "[Y/n]", False: "[y/N]"}
    if default is not None:
        valid[""] = True if default else False
    stdout.write("{} {} ".format(question, default_prompts[default]))
    try:
        return valid[input().lower()]
    except KeyError:
        stdout.write("Valid responses: [y]es/[n]o (case insensitive)\n")
    return yesno(question, default)
